from starling_sim.basemodel.simulation_model import SimulationModel
from starling_sim.basemodel.population.dict_population import DictPopulation
from starling_sim.basemodel.environment.environment import Environment
from starling_sim.models.SB_VS.input import Input
from starling_sim.models.SB_VS.output import Output
from starling_sim.basemodel.schedule.scheduler import Scheduler
from starling_sim.models.SB_VS.user import User
from starling_sim.basemodel.agent.vehicles.station_based_vehicle import StationBasedVehicle
from starling_sim.basemodel.agent.stations.vehicle_sharing_station import VehicleSharingStation


class Model(SimulationModel):
    """
    Simple model for the simulation of a one-way, station-based, vehicle-sharing system
    """

    name = "Station-based vehicle-sharing model"

    agent_type_class = {
        "user": User,
        "vehicle": StationBasedVehicle,
        "station": VehicleSharingStation
    }

    leaving_codes = {
        "FAIL_GET": "Failed to get a vehicle"
    }

    modes = {
        "user": ["walk"],
        "vehicle": ["station"],
        "station": [None, "walk"]
    }

    def __init__(self, parameters):
        """
        Initialisation of the classes used by the station-based model

        :param parameters: SimulationParameters object
        """

        super().__init__(parameters)

        # elements of the model
        self.agentPopulation = DictPopulation(self.agent_type_class.keys())
        self.environment = Environment(self.parameters)

        # inputs and outputs
        self.dynamicInput = Input(self.agent_type_class)
        self.outputFactory = Output()

        # event manager
        self.scheduler = Scheduler()



from starling_sim.models.SB_VS.model import Model as SB_VS_model
from starling_sim.models.SB_VS_R.model import Model as SB_VS_R_model
from starling_sim.models.FF_VS.model import Model as FF_VS_model
from starling_sim.basemodel.parameters.simulation_parameters import parameters_from_file
from starling_sim.utils.paths import model_import_path

import logging
import time
import importlib

models_dict = {
    "SB_VS": SB_VS_model,
    "SB_VS_R": SB_VS_R_model,
    "FF_VS": FF_VS_model
}

model_codes = list(models_dict.keys())


class ModelSimulator:
    """The simulator runs a simulation model"""

    def __init__(self, model):
        """
        Initialize the simulator and its elements

        :param model: SimulationModel object describing the simulation to run
        """

        self.simulationModel = model

    def setup_simulation(self):
        """
        Prepares the simulation model for running
        """

        self.simulationModel.setup()

    def run_simulation(self):
        """
        Call the run method of the simulation model
        """

        self.simulationModel.run()

    def generate_output(self):
        """
        Generates an output of the simulation at its current state
        """

        self.simulationModel.generate_output()

    def init_simulator_from_parameters(simulation_parameters, pkg):
        """
        Returns a simulator initialised according to the model code
        contained in the given parameters

        :param simulation_parameters: SimulationParameters containing "code"
        :param pkg: name of the source package

        :return: ModelSimulator object
        """

        # get the Model class
        model_class = ModelSimulator.get_model_class(simulation_parameters["code"], pkg)

        # create a new instance of the simulation model
        try:
            simulation_model = model_class(simulation_parameters)
        except TypeError as e:
            logging.error("Instantiation of {} failed with message :\n {}"
                          .format(model_class.__name__, e))
            raise e

        # initialise and return the ModelSimulator object
        simulator = ModelSimulator(simulation_model)
        return simulator

    init_simulator_from_parameters = staticmethod(init_simulator_from_parameters)

    def get_model_class(model_code, pkg):
        """
        Get the Model class of the simulation model.

        :param model_code: model code
        :param pkg: name of the source package

        :return: SimulationModel subclass
        """

        # if the code corresponds to a public model, fetch the Model class in models_dict
        if model_code in models_dict:
            model_class = models_dict[model_code]

        # otherwise, import the Model class using importlib
        else:
            model_path = model_import_path(starling_pkg=pkg, model_code=model_code)

            # try to import the module corresponding to the model code
            try:
                module = importlib.import_module(model_path)
            except ModuleNotFoundError:
                raise ModuleNotFoundError("Cannot find the module '{}'.\n"
                                          "    Maybe there is an error in the model code ? "
                                          "Or maybe you forgot to use the -p option of main.py ?".format(model_path))

            # try to get the Model class from the model module
            try:
                model_class = module.Model
            except AttributeError as e:
                logging.error("Cannot find the class Model in {}.models.{}.model".format(pkg, model_code))
                raise e

        return model_class

    get_model_class = staticmethod(get_model_class)


def launch_simulation(parameters_path, pkg):
    """
    Realises the initialisation, setup, run and output of the simulation
    using the given parameters file. Displays logs of execution times

    :param parameters_path: path to parameters files
    :param pkg: name of the source package
    """

    if parameters_path is None:
        raise ValueError("No parameters file provided to simulation launcher.")

    # initialise simulation parameters from parameters file
    simulation_parameters = parameters_from_file(parameters_path)

    # init the simulator
    logging.info("Initializing simulator for the model code "
                 + simulation_parameters["code"] + ", scenario "
                 + simulation_parameters["scenario"] + "\n")
    simulator = ModelSimulator.init_simulator_from_parameters(simulation_parameters, pkg)

    # setup the simulator
    logging.info("Setting entries for: " + simulator.simulationModel.name)
    start = time.time()
    simulator.setup_simulation()
    duration = time.time() - start
    logging.info("End of setup. Elapsed time : {:.2f} seconds\n".format(duration))
    simulator.simulationModel.runSummary["setup_time"] = duration

    # run the simulation
    logging.info("Starting the simulation\n")
    start = time.time()
    simulator.run_simulation()
    duration = time.time() - start
    logging.info("End of simulation run. Elapsed time : {:.2f} seconds\n".format(duration))
    simulator.simulationModel.runSummary["execution_time"] = duration

    shortest_path_count = 0
    for topology in simulator.simulationModel.environment.topologies.values():
        shortest_path_count += topology.shortest_path_count
    logging.info("Number of shortest_path computed : {}".format(shortest_path_count))
    simulator.simulationModel.runSummary["shortest_paths"] = shortest_path_count

    # generate simulation output
    logging.info("Generating outputs of the simulation\n")
    simulator.generate_output()

    logging.shutdown()



from starling_sim.basemodel.agent.agent import Agent
from starling_sim.basemodel.agent.requests import TripRequest, StopPoint
from starling_sim.basemodel.agent.stations.station import Station
from starling_sim.utils.utils import geopandas_polygon_from_points, points_in_zone, json_load, \
    load_schema, validate_against_schema
from starling_sim.utils.paths import gtfs_feeds_folder, schemas_folder
from starling_sim.utils.constants import STOP_POINT_POPULATION, ADD_STOPS_COLUMNS

import pandas as pd


class Operator(Agent):
    """
    Class describing an operator of a transport service.

    Operators manage a fleet of vehicle, and receive requests for the service
    they provide.

    The assignment of these requests is managed by a Dispatcher object.
    """

    OPERATION_PARAMETERS_SCHEMA = {
        "type": "object",
        "title": "Operation parameters",
        "required": [],
        "properties": {
            "routes": {
              "advanced":  True,
              "type": "array",
              "items": {
                "type": "string"
              },
              "title": "GTFS routes to keep in the public transport simulation",
              "description": "List of route ids that should be present in the simulation. If null, keep all routes."
            }
        }
    }

    SCHEMA = {
        "properties": {
            "mode": {
                "type": "object",
                "title": "Operator networks",
                "description": "Networks of the operator's agents",
                "properties": {
                    "fleet": {
                        "title": "Fleet network",
                        "description": "Default network of the operator's fleet",
                        "type": "string"
                    },
                    "staff": {
                        "title": "Staff network",
                        "description": "Default network of the operator's staff",
                        "type": "string"
                    },
                },
                "required": ["fleet"]
            },
            "fleet_dict": {
                "type": "string",
                "title": "Fleet population",
                "description": "Population of the operator's fleet agents"
            },
            "staff_dict": {
                "x-display": "hidden",
                "title": "Staff population",
                "description": "Population of the operator's staff agents",
                "type": "string"
            },
            "zone_polygon": {
                "x-display": "hidden",
                "title": "Service area file",
                "description": "Geojson input file describing the service area",
                "type": "string",
            },
            "depot_points": {
                "advanced": True,
                "title": "Depot points",
                "description": "List of depot points coordinates",
                "type": "array",
                "items": {
                    "type": "array",
                    "items": {
                        "type": "number"
                    },
                    "minItems": 2,
                    "maxItems": 2
                }
            },
            "extend_graph_with_stops": {
                "advanced": True,
                "title": "Extend graph with stops",
                "description": "Extend the graph with service stop points",
                "type": "boolean",
                "default": False
            },
            "operation_parameters": OPERATION_PARAMETERS_SCHEMA,
        },
        "required": ["fleet_dict"],
        "remove_props": ["icon"]
    }

    DISPATCHERS = {}

    @classmethod
    def get_schema(cls):

        schema = cls.compute_schema()

        if len(cls.DISPATCHERS) > 0:

            enum = []

            operation_parameters_schema = schema["properties"]["operation_parameters"]

            for key in cls.DISPATCHERS.keys():

                if "title" in cls.DISPATCHERS[key]:
                    title = cls.DISPATCHERS[key]["title"]
                else:
                    title = key

                dispatchers_ops = dict()
                required_dispatchers_ops = []

                if "online" in cls.DISPATCHERS[key]:
                    online_schema = cls.DISPATCHERS[key]["online"].SCHEMA
                    if isinstance(online_schema, str):
                        online_schema = load_schema(online_schema)
                    dispatchers_ops = {
                        **online_schema["properties"]
                    }
                    if "required" in online_schema:
                        for prop in online_schema["required"]:
                            if prop not in required_dispatchers_ops:
                                required_dispatchers_ops.append(prop)

                if "punctual" in cls.DISPATCHERS[key]:
                    punctual_schema = cls.DISPATCHERS[key]["punctual"].SCHEMA
                    if isinstance(punctual_schema, str):
                        punctual_schema = load_schema(punctual_schema)
                    dispatchers_ops = {
                        **punctual_schema["properties"]
                    }
                    if "required" in punctual_schema:
                        for prop in punctual_schema["required"]:
                            if prop not in required_dispatchers_ops:
                                required_dispatchers_ops.append(prop)

                dispatcher_schema = {
                    "title": title,
                    "properties": {
                        "dispatcher": {
                            "type": "string",
                            "const": key,
                            "title": "Dispatch method"
                        },
                        **dispatchers_ops
                    },
                    "required": required_dispatchers_ops
                }

                enum.append(dispatcher_schema)

            operation_parameters_schema["oneOf"] = enum

        return schema

    @classmethod
    def compute_schema(cls):

        schema = super().compute_schema()

        parent_class = cls.__bases__[0]
        operation_parameters_schema = cls.OPERATION_PARAMETERS_SCHEMA
        if issubclass(parent_class, Operator) \
                and operation_parameters_schema != parent_class.OPERATION_PARAMETERS_SCHEMA:
            if isinstance(operation_parameters_schema, str):
                operation_parameters_schema = load_schema(operation_parameters_schema)
            cls.update_class_schema(schema["properties"]["operation_parameters"], operation_parameters_schema, cls)

        return schema

    def __init__(self, simulation_model, agent_id, fleet_dict, mode=None, staff_dict=None,
                 depot_points=None, zone_polygon=None, operation_parameters=None,
                 extend_graph_with_stops=False, **kwargs):
        """
        Initialise the service operator with the relevant properties.

        The operator class present many different attributes. All are not
        necessarily used when when describing a transport system operator.

        :param simulation_model: SimulationModel object
        :param agent_id: operator's id
        :param fleet_dict: name of the population that contains the operator's fleet
        :param staff_dict: name of the population that contains the operator's staff
        :param depot_points: list of coordinates of the operator depot points
        :param zone_polygon: list of GPS points delimiting the service zone
        :param operation_parameters: additional parameters used for service operation
        :param extend_graph_with_stops: boolean indicating if the graph should be extended with the stop points
        :param kwargs:
        """

        Agent.__init__(self, simulation_model, agent_id, mode=mode, **kwargs)

        # data structures containing the service information

        # parent operator id
        self.parentOperatorId = None

        # parameters determining the service operation
        self.operationParameters = None
        self.init_operation_parameters(operation_parameters)

        # graph extension boolean
        self.extend_graph_with_stops = extend_graph_with_stops

        # set the fleet population (agents that provide the main transport service)
        self.fleet_name = fleet_dict
        self.fleet = self.sim.agentPopulation.new_population(fleet_dict)

        # set the staff population (agents that manage the fleet and operations)
        self.staff_dict_name = staff_dict
        self.staff = self.sim.agentPopulation.new_population(staff_dict)

        # set the dict containing the depot points of the service
        self.depotPoints = dict()
        self.init_depot_points(depot_points)

        # data structure containing the main service structure, if there is one
        self.service_info = None
        self.init_service_info()

        # DataFrame containing the shapes of the service lines with columns
        # stop_id_A	stop_name_A	stop_id_B stop_name_B lon lat sequence distance distance_proportion
        self.line_shapes = None
        self.init_line_shapes()

        # GeoDataFrame representing the service zone of the operator
        self.serviceZone = None
        self.init_zone(zone_polygon)

        # a dict of the service stop points {id: StopPoint}
        self.stopPoints = dict()
        self.init_stops()

        # trip count, used when generating trips
        self.tripCount = 0

        # a dict of service trips {trip_id: [agent, planning]}
        self.trips = dict()
        self.init_trips()

        # requests management

        # request count
        self.requestCount = 0
        # request dict
        self.requests = dict()
        # fulfilled requests
        self.fulfilled = dict()
        # left out requests
        self.leftOutRequests = []
        # minimum horizon for booking a trip
        self.bookingDeadline = 0

        # request dispatcher called punctually
        self.punctual_dispatcher = None

        # dispatcher called to handle requests online
        self.online_dispatcher = None

        self.init_dispatchers()

    # attributes initialisation methods

    def init_operation_parameters(self, operation_parameters):
        """
        Initialise the operation parameters of the operator.

        If the class attribute OPERATION_PARAMETERS_SCHEMA is provided, it is
        used to validate the given parameters and set default values if needed.

        :param operation_parameters: dict of operational parameters
        """
        if operation_parameters is None:
            operation_parameters = dict()
        self.operationParameters = operation_parameters

    def init_service_info(self):
        """
        Import and initialise the data structures containing
        service information and line shapes.
        """

    def init_line_shapes(self):
        """
        Set the service lines shapes file according to the dedicated parameter.
        """

        if "line_shapes_file" in self.operationParameters:
            line_shapes_path = gtfs_feeds_folder() + self.operationParameters["line_shapes_file"]
            self.line_shapes = pd.read_csv(line_shapes_path, sep=";")
            self.line_shapes = self.line_shapes.astype({"stop_id_A": str, "stop_id_B": str})

    def init_zone(self, zone_polygon):
        """
        Set the operator service zone according to the given polygon.

        The service zone is described by a GeoDataFrame containing a
        shapely Polygon.

        :param zone_polygon: list of points describing a polygon
        """

        if isinstance(zone_polygon, str):
            filepath = self.sim.parameters["input_folder"] + zone_polygon
            geojson = json_load(filepath)
            service_zone = geopandas_polygon_from_points(geojson["features"][0]["geometry"]["coordinates"][0])
        elif zone_polygon is None:
            service_zone = None
        else:
            raise TypeError("The zone_polygon parameter must be either an input filename or a list of coordinates")

        self.serviceZone = service_zone

    def init_depot_points(self, depot_points_coord):
        """
        Initialise the depotPoints attribute using the given coordinates.

        :param depot_points_coord:
        :return:
        """

        if depot_points_coord is not None and isinstance(depot_points_coord, list):

            for i, coord in enumerate(depot_points_coord):
                depot_id = "depot-" + str(i)
                if self.mode is None:
                    depot_modes = list(self.sim.environment.topologies.values())
                else:
                    depot_modes = list(self.mode.values())
                position = self.sim.environment.nearest_node_in_modes(coord, depot_modes)
                depot = Station(self.sim, depot_id, position, mode=self.mode["fleet"], agent_type=None)
                self.depotPoints[depot_id] = depot

    def init_stops(self):
        """
        Initialise the stopPoints attribute with using simulation data.
        """

    def add_stops(self, stops_table, id_prefix=""):
        """
        Add the sequence of stops to the stopPoints dict.

        The stops table is a gtfs-like DataFrame, with the columns
        "stop_id" and "stop_name". The method also uses the
        'stop_correspondence_file", which links service stops
        to topology positions.

        :param stops_table: gtfs-like DataFrame
        :param id_prefix: prefix to add to stop ids
        """

        if not set(ADD_STOPS_COLUMNS).issubset(set(stops_table.columns)):
            raise ValueError("Missing columns when adding stop points. Required columns are {} and {} are provided."
                             .format(ADD_STOPS_COLUMNS, stops_table.columns))

        # TODO : this choice of modes is arbitrary, do better
        correspondence_modes = ["walk", self.mode["fleet"]]

        # find the nearest nodes of the stops and extend the graph if asked
        self.sim.environment.add_stops_correspondence(stops_table, correspondence_modes, self.extend_graph_with_stops)

        # browse stops and add StopPoint objects to stopPoints
        for index, row in stops_table.iterrows():

            # get the stop point id, name and position from the table
            stop_point_id = id_prefix + row["stop_id"]
            stop_point_name = row["stop_name"]
            stop_position = row["nearest_node"]

            stop_point_population = self.sim.agentPopulation.new_population(STOP_POINT_POPULATION)

            # if the stop point already exists, get it from the stop point population
            if stop_point_id in stop_point_population:
                stop_point = stop_point_population[stop_point_id]

                # no comparison between the two positions. The stops initialisation must be well ordered

            # otherwise, create a new StopPoint object and add it to the population
            else:
                stop_point = StopPoint(stop_position, stop_point_id, stop_point_name)
                self.sim.agentPopulation.new_agent_in(stop_point, STOP_POINT_POPULATION)

            # add the stop point to the operator stop points dict
            self.stopPoints[stop_point.id] = stop_point

    def init_trips(self):
        """
        Initialise the trips attribute using simulation data.
        """

    def add_trip(self, agent, planning, trip_id=None):
        """
        Add a new trip to the trips dict.

        Associate the trip_id to the agent that realises the trip
        and the planning that describes the trip.

        The default trip id built as <self.id>_T<self.tripCount>.

        :param agent: agent realising the trip, or None if the agent is not known yet
        :param planning: planning of the trip
        :param trip_id: id of the trip, or None

        :return: id of the trip
        """

        # set a default trip id if not provided
        if trip_id is None:
            trip_id = self.id + "_T" + str(self.tripCount)

        # set the trip's agent and planning
        self.trips[trip_id] = [agent, planning]

        # increment trip count
        self.tripCount += 1

        # return trip id
        return trip_id

    def position_in_zone(self, position):
        """
        Test if the given position belongs to the service zone.

        The method returns <self> (the operator) if the position
        is in the service zone, and None otherwise.

        :param position: topology node

        :return: either self (operator), if in zone, or None otherwise
        """

        # if no service zone, all is in service zone
        if self.serviceZone is None:
            return None

        # get position GPS localisation from global environment
        position_localisation = self.sim.environment.get_localisation(position)

        # test if localisation is in service zone
        in_zone = points_in_zone(position_localisation, self.serviceZone)

        # return self if in zone
        if in_zone:
            return self
        else:
            return None

    def init_dispatchers(self):
        """
        Initialise the punctual_dispatcher and online_dispatcher attributes.
        """

        if "dispatcher" not in self.operationParameters:
            return
        else:
            dispatcher = self.operationParameters["dispatcher"]

        if dispatcher not in self.DISPATCHERS:
            raise ValueError("Unsupported operation parameter 'dispatcher' value '{}' (see schema)".format(dispatcher))

        dispatcher_classes = self.DISPATCHERS[dispatcher]

        parameters = self.dispatcher_parameters()

        try:

            if "online" in dispatcher_classes:
                self.online_dispatcher = dispatcher_classes["online"].__new__(dispatcher_classes["online"])
                self.online_dispatcher.__init__(**parameters)

            if "punctual" in dispatcher_classes:
                self.punctual_dispatcher = dispatcher_classes["punctual"].__new__(dispatcher_classes["punctual"])
                self.punctual_dispatcher .__init__(**parameters)

        except (TypeError, KeyError) as e:
            raise ValueError("Instantiation of operator dispatcher failed with message :\n {}".format(str(e)), 30)

    def dispatcher_parameters(self):
        return {
            "simulation_model": self.sim,
            "operator": self,
            "verb": True
        }

    # new requests management

    def new_request(self, agent):
        """
        Create a new request object to the operator, and return it to the requesting agent.

        In the basic Operator's method, the request object is a TripRequest,
        the event is a simple SimPy event, and the request id looks like "R$self.request_count$"

        :param agent: requesting agent
        :return: Request object
        """

        # generate a new request id
        request_id = "R" + str(self.requestCount)

        # initialise a new Request object
        request = TripRequest(agent, self.sim.scheduler.now(), self, request_id)
        request.set_request_event()

        # add it to the request list and increment the request count
        self.requestCount += 1
        self.requests[request_id] = request

        # return the request
        return request

    def assign_request(self, request):
        """
        Assign the given request to an agent of the fleet

        Call the dispatch algorithm in order to assign a vehicle
        to the given request, and signal the schedule change to
        the chosen vehicle.

        :param request:
        """

        if self.online_dispatcher is None:
            self.leftOutRequests.append(request)
        else:
            self.online_dispatcher.online_dispatch(request)

    def cancel_request(self, request):
        """
        Remove the pending queues of the given assigned request.

        For instance, remove the UserStop objects from the stop points lists.

        Called when a request is canceled by the requesting user or the operator.

        :param request: cancelled request
        """

        print("hello")

    def build_trip_request(self, agent=None, origin_position=None, origin_stop=None, origin_time=None,
                           destination_position=None, destination_stop=None, destination_time=None, **kwargs):
        """
        Build a request for a trip with the operator service from the given information.

        Depending on the operator, all parameters may not be necessary, and others mays be added .

        :param agent: agent making the trip request
        :param origin_position: trip origin position in environment
        :param origin_stop: trip origin stop id
        :param origin_time: trip origin time
        :param destination_position: trip destination position in environment
        :param destination_stop: trip destination stop id
        :param destination_time: trip destination time
        :param kwargs: other eventual parameters

        :return: TripRequest completed with trip information set
        """

    # idle service vehicles management

    def idle_behaviour_(self, agent):
        """
        Perform an idle behaviour.

        This method is called when an agent of the
        operator's fleet is idle. The agent then executes
        the behaviour until it is interrupted. The default
        behaviour is to wait indefinitely.

        :param agent: Agent object
        :return: yield a process describing an idle behaviour
        """

        yield agent.execute_process(agent.spend_time_())

    # creation of new service vehicles
    def new_service_vehicle(self, feature):
        """
        Create and initialise a new service vehicle using the model's dynamic input

        :param feature: dictionary containing the service vehicle information
        :return: newly generated ServiceVehicle
        """

        if "operator_id" not in feature["properties"] or feature["properties"]["operator_id"] != self.id:
            self.log_message("Error in the 'operator' field for the generation of "
                             "a service vehicle : {}".format(feature), 30)
            return

        new_service_vehicle = self.sim.dynamicInput.new_agent_input(feature)

        return new_service_vehicle

    # evaluation of a set of user journeys using operator service

    def user_journeys(self, origin, destination, parameters,
                      objective_time, objective_type="start_after"):
        """
        Compute a list of journeys using the operator's transport system.

        Journeys from origin to destination are computed according to the
        operator service, using routers or other itinerary tools.

        The parameters are also used to adjust the journeys characteristics.

        The objective type is either "start_after" or "arrive_before", which correspond
        to the way the journeys will be computed. The objective time specifies the value
        of this start or arrival time.

        A journey is represented as a DataFrame (see journeys.py).

        :param origin: origin position
        :param destination: destination position
        :param parameters: dict of parameters.
        :param objective_time: value of the start or arrival time
        :param objective_type: way of computing journeys, either "start_after" or "arrive_before"
            Default is "start_after".
        :return: a list of journeys represented by DataFrames
        """

    def create_journeys(self, departures_table, arrival_stops, parameters):
        """
        Compute a list of journeys for the given departures timetable and arrival stops.

        :param departures_table:
        :param arrival_stops:
        :param parameters:

        :return: a list of journeys
        """

    def confirm_journey_(self, journey, agent, parameters):
        """
        Confirm the journey choice of the agent and assign its requests if necessary.

        The request assignment may not be immediate, so the confirmation can take some time.

        :param journey:
        :param agent:
        :param parameters:

        :return: boolean indicating if the chosen journey can be executed
        """

        yield self.execute_process(self.spend_time_(0))

        return True

    # utils

    def compute_travel_time_with_detour(direct_travel_time, max_detour):
        """
        Compute the travel time of a trip with a given detour.

        The deviated travel time can be used to compute a maximum acceptable
        travel time, or an estimated travel time.

        If the detour is a float, it is used as a multiplicand with the direct travel time.
        If it is an integer, it is used as a constant (in seconds) added to the direct travel time.

        :param direct_travel_time: direct travel time in seconds
        :param max_detour: maximum detour, either as a multiplicand or as a constant, or None

        :return: deviated travel time (can be a float or an int), or None if max detour is None
        """

        if isinstance(max_detour, float):
            max_travel_time = direct_travel_time * max_detour
        elif isinstance(max_detour, int):
            max_travel_time = direct_travel_time + max_detour
        elif max_detour is None:
            max_travel_time = None
        else:
            raise ValueError("Unsupported type for max detour parameter : {}".format(type(max_detour)))

        return max_travel_time

    compute_travel_time_with_detour = staticmethod(compute_travel_time_with_detour)



"""
This module manages the parameters of the simulation
"""

import logging
import os
from copy import deepcopy

from starling_sim.utils.utils import json_load, validate_against_schema
from starling_sim.utils import paths


class SimulationParameters:
    """
    This class stores the simulation parameters in a dict structure

    It must be extended to build the dict from different sources

    The parameters should be set and accessed using the builtin methods of a dict
    """

    BASE_PARAM_SCHEMA = "parameters.schema.json"

    def __init__(self, parameters, param_schema_path=None):
        """
        This constructor must be extended to initialize the dict from specific arguments
        """

        # get JSON schema
        if param_schema_path is None:
            param_schema_path = paths.schemas_folder() + self.BASE_PARAM_SCHEMA

        self.schema = json_load(param_schema_path)

        # parameters validation
        validate_against_schema(parameters, self.schema)

        # change date format from YYYY-MM-DD to YYYYMMDD
        if "date" in parameters:
            parameters["date"] = parameters["date"].replace("-", "")

        self._parametersDict = parameters

    def __getitem__(self, item):
        """
        Method called when using 'simulationParameters[item]'
        :param item: Name of the parameter accessed
        :return: self._parametersDict[item]
        """

        if item not in self._parametersDict:
            logging.error("Trying to access unknown parameter '" + str(item)
                          + "'. Does it appears in the parameter entry ?")
            raise KeyError("The parameter '" + str(item) + "' does not exist")

        return self._parametersDict[item]

    def __setitem__(self, key, value):
        """
        Method called when using 'simulationParameters[key] = value'
        :param key: Name of the parameter
        :param value: Value associated to the parameter
        :return:
        """

        if key in self._parametersDict:
            logging.warning("Replacing already existing parameter '"
                            + str(key) + "'")

        self._parametersDict[key] = value

    def __contains__(self, item):
        """
        Method called when using 'item in simulationParameters'

        :param item:
        :return: True if item is in _parametersDict, False otherwise
        """

        return item in self._parametersDict

    def keys(self):
        """
        Returns the keys of the _parametersDict attribute

        :return: keys of the _parametersDict attribute
        """

        return self._parametersDict.keys()

    def set_dict(self, params_dict):
        """
        Sets the _parametersDict attribute with the given dict

        :param params_dict:
        :return:
        """

        self._parametersDict = params_dict

    def copy_dict(self):
        """
        Return a deepcopy of the simulation parameters.

        :return: deepcopy of simulation parameters
        """

        return deepcopy(self._parametersDict)


class ParametersFromDict(SimulationParameters):
    """
    Get simulation parameters from an existing dict
    """

    def __init__(self, params_dict):
        """
        Just set the _parametersDict
        :param params_dict:
        """

        super().__init__(params_dict)


class ParametersFromFile(SimulationParameters):
    """
    Get simulation parameters from a json file
    """

    def __init__(self, filepath):
        """
        Initialisation of the param dict by reading params in a json file
        :param filepath:
        """

        # check that the parameters file has the correct name
        if os.path.basename(filepath) != paths.PARAMETERS_FILENAME:
            raise ValueError("The parameters file must be named " + paths.PARAMETERS_FILENAME)

        # read json
        json_param = json_load(filepath)

        # create empty dict
        super().__init__(json_param)


# parameters management

def parameters_from_file(param_path):
    """
    Generates a simulation parameters from the parameters file.

    :param param_path: path to parameters file

    :return: SimulationParameters object
    """

    # create simulation parameters
    parameters = ParametersFromFile(param_path)

    add_paths_to_parameters(parameters)

    return parameters


def add_paths_to_parameters(parameters):
    """
    Adds the input and output paths to the given parameters
    using the 'code' and 'scenario' items
    :param parameters:
    :return:
    """

    model_code = parameters["code"]
    scenario = parameters["scenario"]

    # input folder
    parameters["input_folder"] = paths.scenario_input_folder(model_code, scenario)

    # output folder (will be generated by output)
    parameters["output_folder"] = paths.scenario_output_folder(model_code, scenario)



import random
import logging
import numpy

from starling_sim.basemodel.trace.trace import trace_simulation_end
from starling_sim.utils.utils import import_gtfs_feed
from starling_sim.utils.constants import BASE_LEAVING_CODES
from starling_sim.utils.config import config


class SimulationModel:
    """
    Generic simulation model

    It must be extended to describe concrete systems,
    with their specific topologies and agents
    """

    #: Name of the simulation model
    name = "Unnamed model"

    #: Agent types of the model and their classes
    agent_type_class = None

    #: leaving codes of the model and their description
    leaving_codes = {}

    #: Agent types of the model and their modes
    modes = None

    def __init__(self, parameters):
        """
        Initialisation of the simulation model with instances of its different elements

        This constructor must be extended to instantiate the model elements
        using the correct classes

        :param parameters: SimulationParameters object
        """

        # simulation parameters
        self.parameters = parameters

        # run_summary
        self.runSummary = parameters.copy_dict()

        # add the base leaving codes
        self.add_base_leaving_codes()

        # random seed for the simulation setup and run
        self.randomSeed = parameters["seed"]

        # information to be completed for the specific models

        # elements  of the model
        self.agentPopulation = None
        self.environment = None

        # inputs and outputs
        self.dynamicInput = None
        self.outputFactory = None

        # event manager
        self.scheduler = None

        # complete, unfiltered gtfs timetable, if relevant
        self.gtfs = None

        # TODO : data structure to store simulation information (e.g. demand)
        # self.data = None

    def setup(self):
        """
        Sets the entries of the simulation model and prepares it for the simulation run.
        It is here that the files and data are effectively read and imported in the model.

        This method can be extended to manage and setup other elements of the model
        """

        # set the parameters and initialize the random seed
        self.setup_seeds()

        # start the simulation setup

        logging.info("Simulation environment setup")
        self.environment.setup(self)

        if "gtfs_timetables" in self.parameters:
            logging.info("GTFS tables setup")
            self.setup_gtfs()

        logging.info("Dynamic input setup")
        self.dynamicInput.setup(self)

        logging.info("Output factory setup")
        self.outputFactory.setup(self)

    def run(self):
        """
        Runs the simulation model until the time limit is reached.

        This method can be extended to run other elements of the model
        """

        # if asked, add a process that logs the simulation time every hour
        if "time_log" in self.parameters and self.parameters["time_log"]:
            self.scheduler.new_process(self.periodic_hour_log())

        # create agents and add their loops
        self.scheduler.new_process(self.dynamicInput.play_dynamic_input_())

        # run the simulation
        self.scheduler.run(self.parameters["limit"])

        # trace the end of the simulation for all agents
        trace_simulation_end(self)

    def generate_output(self):
        """
        Generate an output of the current simulation
        """

        self.outputFactory.extract_simulation(self)

    def add_base_leaving_codes(self):
        """
        Add the base leaving codes to the ones specified for the model.

        This will overwrite any custom code named as a base code.
        """

        self.leaving_codes.update(BASE_LEAVING_CODES)

    def setup_seeds(self):
        """
        Set the seeds of the random functions
        """

        random.seed(self.randomSeed)
        numpy.random.seed(self.randomSeed)

    def periodic_hour_log(self):

        while True:

            logging.info("Current simulation time is {}".format(self.scheduler.now()))

            yield self.scheduler.timeout(3600)

    def setup_gtfs(self):
        """
        Setup a gtfs timetable for the simulation.
        """

        # import the gtfs timetable from the zip given in the parameters
        restrict_transfers = config["transfer_restriction"]
        self.gtfs = import_gtfs_feed(self.parameters["gtfs_timetables"], restrict_transfers)

    @classmethod
    def get_agent_type_schemas(cls):

        agent_type_class = cls.agent_type_class
        schemas = dict()
        for agent_type in agent_type_class.keys():
            schemas[agent_type] = agent_type_class[agent_type].get_schema()

        return schemas
